/*	Maze LR2 Level Creator
	by Anthony M Kozar Jr
	Derived from SquareLR.cp
	Begun 6/17/03
	
	5/30/04:   First working version completed.
	2/18/05:   Added #ifdefs for Mac-specific code.
	5/12/05:   Version 1.0
	           Added batch generation of levels, menu system, license info, log file.
	           Added random seeding and fixed several problems with exit placement.
	5/13/05:   Version 1.1
	           Reversed Entrance/Exit to make levels more challenging.
	12/13/05:  Version 1.2
	           Changed to Mona World tiles with invisible walls.
	01/05/12:  Version 1.3
	           Added the ability to choose between "maze styles"
	           Increased the number of maze styles to nine including some with random blocks.
	           Added a function for quickly demoing all styles with the same maze.
 */

#include	<iostream>
#include	<vector>
#include	<string>
#include	<fstream>
#include	<algorithm>
#include	<cstdlib>
#include	<ctime>
#include	<cmath>

#if defined(macintosh) && defined(__MWERKS__)	/* CodeWarrior Mac compiler */
#include	"unix.h"
#include	"SIOUX.h"
#endif

#include	"lr2.h"
#include	"mazegen.h"

using namespace	std;
using	namespace	lr2;

typedef int mstyle_t;


/*	Function Prototypes    */

void	InitConsole();
void	SetTextFileType();
void	SetLevelFileType();
void	PrintLicense();
mstyle_t ChooseStyle();
int	MakeOne();
int	MakeBatchOfMazes();
int	MakeStyleDemos();
coord	FindExit(maze_type<simple_block>* maze);
void	AddEntrance( maze_type<simple_block>* maze );
void	WriteLevel(ofstream &fileout, maze_type<simple_block>* maze, mstyle_t style);
void	Header(ofstream &fileout, world_t world, int16 backgndID = (int16)0x2af8);
void	Monks(ofstream &fileout);
void	Items(ofstream &fileout);
void	MazeItems(ofstream &fileout, maze_type<simple_block>* maze);
void	MazeLayout(ofstream &fileout, maze_type<simple_block>* maze, mstyle_t style);
void	MazeLayout2(ofstream &fileout, maze_type<simple_block>* maze, mstyle_t style);
void	Gold(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor);
void	EntryPoint(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor);
void	WriteCoord(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor);
void	YHeader(ofstream &fileout, int16 Xnum, int16 Ycor);
void	ZHeader(ofstream &fileout, int16 Ynum, int16 Zcor);
void	Block(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor, /*block_t*/int16 theBlock = kBCBuilderBlue);
void	InvisibleBlock(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor);
void	ExitBlock(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor, world_t world = kBuilderWorld);
void	SqrItems(ofstream &fileout, int sqr_size);
void	SquareLayout(ofstream &fileout, int sqr_size);
//void	Divider(ofstream &fileout, int length);
//void	Terminator(ofstream &fileout, int length);

/*	Constants   */

const	int	PLATFORM_Z		= 100;
const	int	MAZE_Z		= 99;
const	int	ENTRY_Z		= 95;

typedef int mstyle_t;

const mstyle_t	MIN_STYLE			= 0;
const mstyle_t	MAX_STYLE			= 9;

// different styles for brick choices
const mstyle_t	kRandomStyle		= 0;
const mstyle_t	kBuilderWalled		= 1;
const mstyle_t	kMonaStoneWalkways	= 2;
const mstyle_t	kJungleStoneWalkways	= 3;
const mstyle_t	kGearTranslucent		= 4;
const mstyle_t	kWackyLemonSlices		= 5;
const mstyle_t	kWackyLimeSlices		= 6;
const mstyle_t	kWackyOrangeSlices	= 7;
const mstyle_t	kWackyCitrusMix		= 8;
const mstyle_t	kJungleVegatatedWalks	= 9;
// be sure to increase MAX_STYLE above if you add more!

const string  STYLE_NAMES[MAX_STYLE+1] = {
			"Use a random style",
			"Walled Builder world maze",
			"Mona world stone tiles",
			"Jungle world stone tiles",
			"Translucent walkways",
			"Lemon slices",
			"Lime slices",
			"Orange slices",
			"Citrus mix",
			"Jungle stone with vegatation"
			};

const world_t STYLE_WORLD_CODES[MAX_STYLE+1] = { kBuilderWorld, kBuilderWorld, kMonaWorld, kJungleWorld, 
								 kGearWorld, kWackyWorld, kWackyWorld, kWackyWorld, 
								 kWackyWorld, kJungleWorld };

const	char	VERSION[]		= "1.3";

const	char	LICENSE_TEXT[]	=
	"Maze Generator\n"
	"Copyright (c) 2005, 2012 Anthony M. Kozar Jr.\n"
	"\n"
	"Redistribution and use in source and binary forms, with or without\n"
	"modification, are permitted provided that the following conditions are met:\n"
	"\n"
	"1.  Redistributions of source code must retain the above copyright\n"
	"    notice, this list of conditions and the following disclaimer. \n"
	"2.  Redistributions in binary form must reproduce the above copyright\n"
	"    notice, this list of conditions and the following disclaimer in the\n"
	"    documentation and/or other materials provided with the distribution. \n"
	"3.  The name of the author may not be used to endorse or promote\n"
	"    products derived from this software without specific prior written\n"
	"    permission. \n"
	"\n"
	"THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED\n"
	"WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF\n"
	"MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO\n"
	"EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,\n"
	"SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,\n"
	"PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;\n"
	"OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,\n"
	"WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR\n"
	"OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF\n"
	"ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n"
	"\n"
	"Maze Generator is not made by or supported by GT Interactive Software\n"
	"Corp., The WizardWorks Group, Inc., or any of such parties affiliates and\n"
	"subsidiaries.\n"
	"\n";

/*	Main Program    */

int	main()
{
	int	option;
	
	InitConsole();
	srand(time(NULL));
	
	cout	<< "Maze Generator " << VERSION << endl;
	cout	<< "by Anthony M. Kozar Jr." << endl;
	
	do	{
		cout	<< endl;
		cout	<< "Choose an option:" << endl
			<< "  (1)  Make one maze level" << endl
			<< "  (2)  Make a set of maze levels" << endl
			<< "  (3)  Print license information" << endl 
			<< "  (0)  Quit" << endl << "? ";
		cin	>> option;
		cout	<< endl;

		switch (option)	{
			case 0: break;
			case 1:
				if	(MakeOne())			option = 0;	// quit if make fails
				break;
			case 2:
				if	(MakeBatchOfMazes())	option = 0;	// quit if make fails
				break;
			case 3:
				PrintLicense();
				break;
			case 99:							// run tests and demos
				if	(MakeStyleDemos())	option = 0;	// quit if make fails
				break;
			default:
				cout	<< "Please select one of the available choices.\n";
				break;
		}
	}
	while	(option && cin.good());
	
	cout	<< "Goodbye." << endl;
	return 0;
}

void	InitConsole()
{
#if defined(macintosh) && defined(__MWERKS__)	/* CodeWarrior Mac compiler */
	// don't prompt to save console window
	SIOUXSettings.asktosaveonclose = false;
	SIOUXSettings.rows = 48;
#endif
	
	SetLevelFileType();
	return;
}

void	SetTextFileType()
{
#if defined(macintosh) && defined(__MWERKS__)	/* CodeWarrior Mac compiler */
	// set output file type to LR2 level
	_fcreator = 'R*ch';
	_ftype = 'TEXT';
#endif
	
	return;
}

void	SetLevelFileType()
{
#if defined(macintosh) && defined(__MWERKS__)	/* CodeWarrior Mac compiler */
	// set output file type to LR2 level
	_fcreator = 'LR2E';
	_ftype = 'LEVL';
#endif
	
	return;
}


void	PrintLicense()
{
	cout	<< LICENSE_TEXT;
	return;
}

mstyle_t ChooseStyle()
{
	mstyle_t	style;
	
	do	{
		cout	<< endl;
		cout	<< "Choose a maze style:" << endl;
		for (int i = MIN_STYLE; i <= MAX_STYLE; i++) {
			cout << "  (" << i << ")  " << STYLE_NAMES[i] << endl;
		}
		cout	<< "? ";
		cin	>> style;
		cout	<< endl;

		if (style < MIN_STYLE || style > MAX_STYLE) {
			cout << "Please choose from the available styles!" << endl;
		}
		else return style;
	}
	while	(cin.good());

	return kRandomStyle;
}

int	MakeOne()
{
	string filename;
	ofstream fout;
	maze_type<simple_block>* maze = NULL;
	mstyle_t	style;
	
	cout	<< "Column and Row sizes must be odd integers within the range 3 to 200.\n\n";
	maze = MakeOneMaze();
	AddEntrance(maze);
	maze->print_maze(cout);
	
	style = ChooseStyle();
	
	cout	<< "Enter a filename to save as: " << endl;
	cin	>> filename;

	fout.open(filename.c_str(), ios_base::out | ios_base::binary); 	// open file for binary output
	if (! fout)	{
		cout	<< "Could not open the output file.  Exiting." << endl;
		return -1;
	}
	
	WriteLevel(fout, maze, style);
	fout.close();
	return 0;
}

int	MakeStyleDemos()
{
	const	int rows = 13, columns = 15;
	int	seed = 27;
	int	filenamelen;
	string foldername;
	char	pathname[255];
	ofstream fout;
	maze_type<simple_block>* maze = NULL;
	mstyle_t	style;
	
	// format filenames differently for Mac and Windows
#if defined(macintosh) && defined(__MWERKS__)
	char	formatstr[] = ":%s:%d - %s\0";
	const int MAXLEN = 27;
#else
	char	formatstr[] = "%s/%d_%s\0";
	const int MAXLEN = 8;
#endif
	
	// seed random number gen
	// cout	<< "Enter a seed value for the random number generator: ";
	// cin	>> seed;
	srand(seed);

	// create the maze
	maze = GenerateMaze(rows, columns);
	AddEntrance(maze);
	maze->print_maze(cout);
	
	cout	<< endl << "Enter a folder name to write demos to (folder must already exist)." << endl;
	cin	>> foldername;

	for (style = (MIN_STYLE+1); style <= MAX_STYLE; style++) {
		// pathname = foldername + style # + style name + .lvl
		sprintf(pathname, formatstr, foldername.c_str(), style, STYLE_NAMES[style].c_str());
		if (strlen(pathname) - foldername.size() > MAXLEN)
			strcpy(&(pathname[MAXLEN + foldername.size()]), ".lvl");	// truncate and add extension
		else	strcat(pathname, ".lvl");						// just add extension
		fout.open(pathname, ios_base::out | ios_base::binary); 		// open file for binary output
		if (! fout)	{
			cout	<< "Could not open the output file  " << pathname << endl;
			return -1;
		}
		
		WriteLevel(fout, maze, style);
		fout.close();
	}
	
	cout << "Finished writing demo levels." << endl;
	return 0;
}


int	MakeBatchOfMazes()
{
	int	numlevels, minrow, mincol, maxrow, maxcol, rows, columns, randomly, seed, digits;
	string fileprefix;
	char	filename[255];
	ofstream logfile;
	maze_type<simple_block>* maze = NULL;
	mstyle_t	style;
	
	cout	<< "Column and Row sizes must be odd integers within the range 3 to 200.\n\n";
	cout	<< "How many maze levels? ";
	cin	>> numlevels;
	cout	<< "Minimum number of rows? ";
	cin	>> minrow;
	cout	<< "Maximum number of rows? ";
	cin	>> maxrow;
	cout	<< "Minimum number of columns? ";
	cin	>> mincol;
	cout	<< "Maximum number of columns? ";
	cin	>> maxcol;
	
	// check for validity of entered values
	if	(minrow < 0 or mincol < 0)	{
		cout	<< "Row and column sizes must be greater than zero!\n";
		return 0;
	}
	if	(maxrow < minrow or maxcol < mincol) {
		cout	<< "Maximum values cannot be smaller than minimum values!\n";
		return 0;
	}
	
	cout	<< "(1)  Randomly" << endl
		<< "(2)  Increase from minimum to maximum" << endl
		<< "How should the maze size be determined? ";
	cin	>> randomly;
	randomly = ! (randomly - 1);
	
	style = ChooseStyle();

	cout	<< "Enter a filename prefix: ";
	cin	>> fileprefix;
	
	cout	<< "Enter a seed value for the random number generator (0 for clock): ";
	cin	>> seed;
	
	if	(!seed)	seed = time(NULL);
	srand(seed);
	
	// open the log file and write parameters
	SetTextFileType();
	sprintf(filename, "%s%s\0", fileprefix.c_str(), ".log");
	logfile.open(filename);
	if (!logfile)	{
		cout	<< "Could not open log file" << filename << ".  Exiting." << endl;
		return -1;
	}
	logfile << filename << endl 
		  << "Levels created by Maze Generator" << endl
		  << "Version " << VERSION << endl << endl
		  << "Generating " << numlevels << " maze levels." << endl
		  << "Minimum rows:    " << minrow << endl
		  << "Maximum rows:    " << maxrow << endl
		  << "Minimum columns: " << mincol << endl
		  << "Maximum columns: " << maxcol << endl
		  << "Random seed:     " << seed   << endl << endl;
	
	SetLevelFileType();
	digits = (int)log10((double)numlevels) + 1;
	for	(int i = 1; i <= numlevels; i++)	{
		// calculate size
		if	(randomly)	{
			if	(minrow == maxrow) rows = minrow;
			else	rows = (rand() % (maxrow-minrow)) + minrow;
			if	(mincol == maxcol) columns = mincol;
			else	columns = (rand() % (maxcol-mincol)) + mincol;
		}
		else	{	// linearly interpolate between min and max
			rows = (((float)i/(float)numlevels) * (maxrow-minrow)) + minrow;
			columns = (((float)i/(float)numlevels) * (maxcol-mincol)) + mincol;
		}
		// make sure rows and columns are odd
		if	(!(rows%2))		++rows;
		if	(!(columns%2))	++columns;
		
		// create the maze
		maze = GenerateMaze(rows, columns);
		AddEntrance(maze);
		
		//cout	<< endl;
		//maze->print_maze(cout);					// print the maze to standard out
		//cout	<< endl;
		
		// determine filename, open, and write to file
		//string filename;
		ofstream fout;
		
		//filename = fileprefix + string(columns) + "x" + string(rows);
		sprintf(filename, "%s%.*d-%dx%d%s\0", fileprefix.c_str(), digits, i, columns, rows, ".lvl");
		fout.open(filename, ios_base::out | ios_base::binary); 	// open file for binary output
		if (! fout)	{
			cout	<< "Could not open output file" << filename << ".  Exiting." << endl;
			return -1;
		}
		
		cout	<< "Writing " << columns << "x" << rows << " maze to file " << filename << ".\n";
		logfile << "Writing " << columns << "x" << rows << " maze to file " << filename << ".\n\n";
		maze->print_maze(logfile);					// print the maze to log file
		logfile << endl;
		
		WriteLevel(fout, maze, style);
		fout.close();
		delete maze;
		maze = NULL;
	}

	logfile.close();
	return 0;
}	

coord	FindExit(maze_type<simple_block>* maze)
{
	coord	loc;
	int	xsize = maze->xsize,
		ysize = maze->ysize;
		
	for	(loc.second = 1; loc.second <= ysize; loc.second++) {
		for	(loc.first = 1; loc.first <= xsize; loc.first++)	{
			if	(maze->block_type(coord(loc.first, loc.second)) == kMExit)	{
				return loc;
			}
		}
	}
	// should not get to this point
	cerr	<< "No exit block in maze!" << endl;
	exit(-1);
	
	return loc;
}

void	AddEntrance( maze_type<simple_block>* maze )
{
	coord	start, exit;
	bool	again;
	int	columns = maze->xsize;
	int	rows = maze->ysize;
	
	exit = FindExit(maze);
	// choose where to put the entrance
	do	{
		again = false;
		int	edge = rand() % 2;		// choose an edge of the maze
		switch (edge)	{
			case	0:				// left column
				start.first = 1;
				start.second = 2 * (rand()%(rows/2) + 1);	// only even numbered spots
				break;
			case	1:				// bottom row
				start.first = 2 * (rand()%(columns/2) + 1);
				start.second = rows;
				break;
			/* only want the entrance on the above edges */
			/*case	2:				// right column
				start.first = columns;
				start.second = 2 * (rand()%(rows/2) + 1);
				break;
			case	3:				// or top row
				start.first = 2 * (rand()%(columns/2) + 1);
				start.second = 1;
				break;*/
		}
		if	(start.first == exit.first and start.second == exit.second)
			again = true;			// can't have start and exit in the same spot
	}
	while	(again);
	
	maze->block_type(start, kMStart);
	return;	
}

void	WriteLevel(ofstream &fileout, maze_type<simple_block>* maze, mstyle_t style)
{
	//cout  << "writing ... ";
	if (style == kRandomStyle) {
		style = (rand() % MAX_STYLE) + 1;	// choose between 1 and MAX_STYLE
	}
	
	Header( fileout, STYLE_WORLD_CODES[style] );
	Monks(fileout);
	MazeItems(fileout,maze);
	//Divider(fileout,180);
	switch (style) {
		case kBuilderWalled:	MazeLayout(fileout,maze,style); break;
		default: 			MazeLayout2(fileout,maze,style); break;
	}
	
	//Terminator(fileout,222);
	//cout  << "done." << endl;
	
	return;
}

void	Header(ofstream &fileout, world_t world, int16 backgndID)
{
	//using	namespace	lr2;
	
	//cout  << "header, ";
	//fileout << LEVEL_FILE_ID;			// "magic" file ID
	fileout << int16(0xD87E) << int16(0x1B17);
	
	// I don't really know the purpose of the first part of the header
	if	( world == kWackyWorld )		// Wacky world header is different
		for	(int i = 0; i < H1_SIZE; i++)	fileout << WACKY_HEADER[i];
	else	for	(int i = 0; i < H1_SIZE; i++)	fileout << HEADER_PART_1[i];
	
	// values for the world and background of the level
	fileout << world;
	fileout << (int16)0x0000;
	fileout << backgndID;
	
	// the end of the header is all zeros unless scroll-locking is enabled
	for	(int i = 0; i < H2_SIZE; i++)	fileout << HEADER_PART_2[i];
	return;
}

void	Monks(ofstream &fileout)
{
	//using	namespace	lr2;
	
	//cout  << "no monks, ";
//	for	(int i = 0; i < 4; i++)		fileout << monk_header[i];
	fileout << int16(0x0000);
	fileout << int16(0x0005);
	return;
}

void	Items(ofstream &fileout)
{
	cout  << "items, ";
	/*restore item_data
	for i = 1 to 32
	read value
	fileout << (char)value;
	next i
	*/	
	return;
}

void	MazeItems(ofstream &fileout, maze_type<simple_block>* maze)
{
	//cout  << "items, ";

	// item block header
	fileout << int16(1);							// one item (the LR entry)
	fileout << int16(3);							// identifies this as beginning of items

	// find entry point
	int	xsize = maze->xsize,
		ysize = maze->ysize;
		
	for	(int y = kMinimumY; y <= (ysize+kMinimumY-1); y++) {
		for	(int x = (xsize+kMinimumX-1); x >= kMinimumX; x--)	{
			if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMStart)	{
				EntryPoint(fileout, x, y, ENTRY_Z);			// add it
				goto done;
			}
		}
	}
	
	done:
	return;
}

/*    MazeLayout()

	This layout algorithm works for the original Builder World style mazes.
	I.E.  it creates a square platform with visible walls above to create the maze.
 */
void	MazeLayout(ofstream &fileout, maze_type<simple_block>* maze, mstyle_t style)
{
	//cout  << "layout, ";

	int	currentZ = kMinimumZ;
	for	( ; currentZ < MAZE_Z; currentZ++ )
		fileout << int16(0xFFFF);						// "skip this Z plane"
		
	// add maze blocks
	int	xsize = maze->xsize,
		ysize = maze->ysize,	
		xnum;
		
	ZHeader(fileout, ysize, currentZ);
	for	(int y = kMinimumY; y <= (ysize+kMinimumY-1); y++) {
		xnum = 0;
		for	(int x = (xsize+kMinimumX-1); x >= kMinimumX; x--) {	// count the blocks for this Y
			if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMWall)	++xnum;
			if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMExit)	++xnum;
		}
		YHeader(fileout, xnum, y);
		for	(int x = (xsize+kMinimumX-1); x >= kMinimumX; x--) {	// add the blocks for this Y
			if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMWall)
				Block(fileout,x,y,currentZ);				// use default block type
			else	if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMExit)
				ExitBlock(fileout,x,y,currentZ);			// exit block
		}
	}
	++currentZ;
	
	// make platfrom - a rectangle of blocks
	ZHeader(fileout, ysize, currentZ);
	for	(int y = kMinimumY; y <= (ysize+kMinimumY-1); y++) {
		YHeader(fileout, xsize, y);
		for	(int x = (xsize+kMinimumX-1); x >= kMinimumX; x--)	{
			Block(fileout,x,y,currentZ);					// use default block type
		}
	}
	++currentZ;
	
	for	( ; currentZ <= kMaximumZ; currentZ++ )
		fileout << int16(0xFFFF);						// "skip this Z plane"
				
	return;
}

/*    MazeLayout2()

	This layout algorithm works for the "flat, open"-style mazes.
	I.E.  it creates walkways in open space with no continous floor or platform
	and with invisible walls above to keep the player from falling.
 */
void	MazeLayout2(ofstream &fileout, maze_type<simple_block>* maze, mstyle_t style)
{
	//cout  << "layout, ";

	int	currentZ = kMinimumZ;
	for	( ; currentZ < MAZE_Z; currentZ++ )
		fileout << int16(0xFFFF);						// "skip this Z plane"
		
	// add maze blocks
	int	xsize = maze->xsize,
		ysize = maze->ysize,	
		xnum, yactual, ycount;
		
	// invisible walls and exit in this layer
	yactual = 0;
	for	(int y = kMinimumY; y <= (ysize+kMinimumY-1); y++)		// count Y's that contain blocks
		for	(int x = (xsize+kMinimumX-1); x >= kMinimumX; x--)
			if	((maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMWall) ||
				 (maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMExit)) {
				++yactual;
				break;
			}
	ZHeader(fileout, yactual, currentZ);
	ycount = 0;
	for	(int y = kMinimumY; y <= (ysize+kMinimumY-1); y++) {
		xnum = 0;
		for	(int x = (xsize+kMinimumX-1); x >= kMinimumX; x--) {	// count the blocks for this Y
			if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMWall)	++xnum;
			if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMExit)	++xnum;
		}
		if	(xnum) {
			++ycount;
			YHeader(fileout, xnum, y);
			for	(int x = (xsize+kMinimumX-1); x >= kMinimumX; x--) {	// add the blocks for this Y
				if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMWall)
					InvisibleBlock(fileout,x,y,currentZ);		// invisible block
				else
				if	(maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMExit)
					ExitBlock(fileout,x,y,currentZ,STYLE_WORLD_CODES[style]);	// exit block
			}
		}
	}
	if (ycount != yactual)	{
		cerr	<< "ycount (" << ycount << ") did not equal yactual (" << yactual 
			<< ") for exit/wall layer" << endl;
		exit(-1);
	}
	++currentZ;
	
	// determine block code(s) for this maze style
	int16			blockcode = int16(0);
	vector<int16>	bcodes;
	bool			chooseRandom = false;
	int			choice, numchoices;

	switch (style) {	
		case kRandomStyle:		blockcode = kBCBuilderBrown; break;	// shouldn't see this case
		case kBuilderWalled:		blockcode = kBCBuilderBlue; break;
		case kMonaStoneWalkways:	blockcode = kBCMonaFlatStone; break;
		case kJungleStoneWalkways:	blockcode = kBCJungleFlatStone; break;
		case kGearTranslucent:		blockcode = kBCGearTranslucent; break;
		case kWackyLemonSlices:		blockcode = kBCLemonSlice; break;
		case kWackyLimeSlices:		blockcode = kBCLimeSlice; break;
		case kWackyOrangeSlices:	blockcode = kBCOrangeSlice; break;
		case kWackyCitrusMix:		chooseRandom = true;
							bcodes.push_back(kBCLemonSlice);
							bcodes.push_back(kBCLimeSlice);
							bcodes.push_back(kBCOrangeSlice);
							break;
		case kJungleVegatatedWalks:	chooseRandom = true;
							bcodes.push_back(kBCJungleFlatStone);
							bcodes.push_back(kBCJungleFlatStone); // weight plain block 50%
							bcodes.push_back(kBCJungleStoneVeg1);
							bcodes.push_back(kBCJungleStoneVeg2);
							break;
		default:				blockcode = kBCBuilderBrown; break;
	}
	
	numchoices = bcodes.size();
	
	// make floor of the maze
	yactual = 0;
	for	(int y = kMinimumY; y <= (ysize+kMinimumY-1); y++)		// count Y's that contain blocks
		for	(int x = (xsize+kMinimumX-1); x >= kMinimumX; x--)
			if	((maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMEmpty) ||
				 (maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMStart) ||
				 (maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMExit)) {
				++yactual;
				break;
			}
	ZHeader(fileout, yactual, currentZ);
	ycount = 0;
	for	(int y = kMinimumY; y <= (ysize+kMinimumY-1); y++) {
		xnum = 0;
		for	(int x = (xsize+kMinimumX-1); x >= kMinimumX; x--) {	// count the blocks for this Y
			if	((maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMEmpty) ||
				 (maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMStart) ||
				 (maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMExit))
				++xnum;
		}
		if	(xnum) {
			++ycount;
			YHeader(fileout, xnum, y);
			for	(int x = (xsize+kMinimumX-1); x >= kMinimumX; x--)	{
				if	((maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMEmpty) ||
					 (maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMStart) ||
					 (maze->block_type(coord( x-(kMinimumX-1), y-(kMinimumY-1) )) == kMExit))
					if (!chooseRandom)  Block(fileout,x,y,currentZ,blockcode);
					else {
						choice = rand() % numchoices;
						Block(fileout,x,y,currentZ,bcodes[choice]);
					}
			}
		}
	}
	if (ycount != yactual)	{
		cerr	<< "ycount (" << ycount << ") did not equal yactual (" << yactual 
			<< ") for maze floor layer" << endl;
		exit(-1);
	}
	++currentZ;
	
	for	( ; currentZ <= kMaximumZ; currentZ++ )
		fileout << int16(0xFFFF);						// "skip this Z plane"
				
	return;
}


inline void	EntryPoint(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor)
{
	fileout << int16(0x00A2);					// entry point sequence
	fileout << int16(0x0003);
	fileout << int16(0xFFFF);
	WriteCoord(fileout, Xcor, Ycor, Zcor);
	return;
}

inline void	Gold(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor)
{
	fileout << int16(0x00AD);					// gold piece sequence
	fileout << int16(0x0001);
	fileout << int16(0xFFFF);
	WriteCoord(fileout, Xcor, Ycor, Zcor);
	return;
}

inline void	ZHeader(ofstream &fileout, int16 Ynum, int16 Zcor)
{
	//cout  << "   Z";

	fileout << Ynum;
	fileout << int16(0x0003);
	fileout << Zcor;
	return;
}

inline void	YHeader(ofstream &fileout, int16 Xnum, int16 Ycor)
{
	//cout  << " Y";
	
	fileout << Xnum;
	fileout << int16(0x000A);
	fileout << Ycor;
	return;
}

void	Block(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor, /*block_t*/ int16 theBlock)
{
	//cout  << "+";

	for	(int i = 0; i < B1_SIZE; i++)	fileout << DIGGABLE_BLOCK1[i];
	WriteCoord(fileout,Xcor,Ycor,Zcor);
	for	(int i = 0; i < B2_SIZE; i++)	fileout << DIGGABLE_BLOCK2[i];
	fileout << theBlock;

	return;
}

void	InvisibleBlock(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor)
{
	//cout  << "+";

	for	(int i = 0; i < INV1_SIZE; i++)	fileout << INVISIBLE_BLOCK1[i];
	WriteCoord(fileout,Xcor,Ycor,Zcor);
	for	(int i = 0; i < INV2_SIZE; i++)	fileout << INVISIBLE_BLOCK2[i];

	return;
}

void	ExitBlock(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor, world_t world)
{
	unsigned short	w = world;
	//cout  << "E";

	for	(int i = 0; i < E1_SIZE; i++)	fileout << EXIT_BLOCK1[i];
	WriteCoord(fileout,Xcor,Ycor,Zcor);
	for	(int i = 0; i < E2_SIZE; i++)	fileout << EXIT_BLOCK2[i];
	switch (w) {
		case 0:					// kMonaWorld
			fileout << int16(kBCMonaExit);
			break;
		case 1:					// kWackyWorld
			fileout << int16(kBCWackyExit);
			break;
		case 2:					// kIndustyWorld
			fileout << int16(kBCIndustryExit);
			break;
		case 4:					// kJungleWorld
			fileout << int16(kBCJungleExit);
			break;
		case 6:					// kGearWorld
			fileout << int16(kBCGearExit);
			break;
		case 9:					// kCreditsWorld
			fileout << int16(kBCCreditsExit);
			break;
		case 5:					// kBuilderWorld
		case 8:					// kHubWorld (which exit should we use?)
		default:
			fileout << int16(kBCBuilderExit);
			break;
	}
	return;
}

inline void	WriteCoord(ofstream &fileout, int16 Xcor, int16 Ycor, int16 Zcor)
{
	fileout << Xcor;
	fileout << Ycor;
	fileout << Zcor;
	return;
}
		
				

void	SqrItems(ofstream &fileout, int sqr_size)
{
	cout  << "items, ";
	
	// item block header
	fileout << int16(5);						// five items total
	fileout << int16(3);						// identifies this as beginning of items
	
	Gold(fileout, sqr_size+99, 100, 99);			// four gold pieces at corners
	Gold(fileout, 100, 100, 99);
	Gold(fileout, sqr_size+99, sqr_size+99, 99);
	Gold(fileout, 100, sqr_size+99, 99);
	EntryPoint(fileout, int(100 + sqr_size/2), 100, 97); 	// center on one edge of square
	
	return;
}

void	SquareLayout(ofstream &fileout, int sqr_size)
{
	int	sqr_center;
	cout  << "layout, ";
	
	// create an exit block in center of Z=99
	ZHeader(fileout, 1, 99);
	sqr_center = 100 + sqr_size/2;
	YHeader(fileout, 1, sqr_center);
	ExitBlock(fileout, sqr_center, sqr_center, int16(99));	// center it
	
	// make a square of blocks sqr_size on each side
	ZHeader(fileout,sqr_size, 100);
	for	(int y = 100; y <= (100 + sqr_size - 1); y++) {
		YHeader(fileout,sqr_size, y);
		for	(int x = (100 + sqr_size - 1); x >= 100; x--) {
			Block(fileout,x,y,100);					// use default block type
		}
	}
	
	return;
}

/*void	Divider(ofstream &fileout, int length)
{
	cout  << "divider, ";
	
	for	( int i = 0; i < length; i++ )
		fileout << (int8)255;

	return;
}

void	Terminator(ofstream &fileout, int length)
{
	cout  << "terminator. ";
	
	for	( int i = 0; i < length; i++ )
		fileout << (int8)255;

	return;
}

*/